# parse_options: {"target-version": "3.15"}
[*x for x in y]
{*x for x in y}
{**x for x in y}
(*x for x in y)
f(*x for x in y)
[*x async for x in y]
{*x async for x in y}
{**x async for x in y}
(*x async for x in y)
