mod ann_eof;
mod ann_insertion;
mod ann_multiline;
mod ann_multiline2;
mod ann_removed_nl;
mod ensure_emoji_highlight_width;
mod first_snippet_is_primary;
mod fold_ann_multiline;
mod fold_bad_origin_line;
mod fold_leading;
mod fold_trailing;
mod highlight_diff_line_with_wide_characters;
mod highlight_duplicated_diff_lines;
mod highlight_first_line_tab_371;
mod highlight_source;
mod highlight_source_multi_width_chars;
mod highlight_source_zero_width_chars;
mod issue_9;
mod multiline_removal_indent;
mod multiline_removal_last_line_tabs;
mod multiline_removal_suggestion;
mod multiple_annotations;
mod multiple_highlight_duplicated;
mod multiple_multiline_removal;
mod primary_title_second_group;
mod regression_leading_tab_label_alignment;
mod regression_leading_tab_long_line;
mod simple;
mod strip_line;
mod strip_line_char;
mod strip_line_non_ws;
mod styled_title;
