# flags: --line-ranges=5-5 --line-ranges=9-9 --line-ranges=13-13
# NOTE: If you need to modify this file, pay special attention to the --line-ranges=
# flag above as it's formatting specifically these lines.
alist   =   [
        1, 2
]

adict   =   {
    "key"  :  "value"
}

func_call  (
    arg  = value
)
