'single' # this string is treated as a docstring
"double"
r'r single'
r"r double"
f'f single'
f"f double"
fr'fr single'
fr"fr double"
rf'rf single'
rf"rf double"
b'b single'
b"b double"
rb'rb single'
rb"rb double"
br'br single'
br"br double"

'''single triple'''
"""double triple"""
r'''r single triple'''
r"""r double triple"""
f'''f single triple'''
f"""f double triple"""
fr'''fr single triple'''
fr"""fr double triple"""
rf'''rf single triple'''
rf"""rf double triple"""
b'''b single triple'''
b"""b double triple"""
rb'''rb single triple'''
rb"""rb double triple"""
br'''br single triple'''
br"""br double triple"""

'single1' 'single2'
'single1' "double2"
"double1" 'single2'
"double1" "double2"

def docstring_single_triple():
    '''single triple'''

def docstring_double_triple():
    """double triple"""

def docstring_double():
    "double triple"

def docstring_single():
    'single'
