if 1: 10
elif 2: 20
else: 30

if True:
    1
    ...
if x < 1:
    ...
else:
    pass

if a:
    pass
elif b:
    ...

if a and b:
    ...
elif True:
    ...
elif c:
    ...
elif d:
    ...
else:
    f()

# Valid test expression
if a := b: ...
elif a := b: ...
if lambda x: x: ...
elif lambda x: x: ...
if await x: ...
elif await x: ...
if (yield x): ...
elif (yield x): ...
